from abc import ABC, abstractmethod


class SerializableProtocolMessage(ABC):
    @abstractmethod
    def dump(self) -> bytes:
        """
        Dump class protocol message to bytes
        :return: NATS protocol message as bytes-encoded string
        """
        raise NotImplementedError()



import inspect
from typing import Callable, Any

from renats.client.message import Message

SubscriptionCallbackType = Callable[[Message], Any]


class Subscription:
    def __init__(self, sid: str, subject: str, callback: SubscriptionCallbackType):
        self.id = sid
        self.subject = subject
        self.callback = callback

    async def __call__(self, message: Message):
        if inspect.iscoroutinefunction(self.callback):
            await self.callback(message)
        else:
            self.callback(message)



from typing import Final

from renats.protocol import utils

HEADERS_VERSION: Final[bytes] = b"NATS/1.0"

INFO: Final[bytes] = b"INFO"
CONNECT: Final[bytes] = b"CONNECT"
PUB: Final[bytes] = b"PUB"
HPUB: Final[bytes] = b"HPUB"
SUB: Final[bytes] = b"SUB"
UNSUB: Final[bytes] = b"UNSUB"
MSG: Final[bytes] = b"MSG"
HMSG: Final[bytes] = b"HMSG"
PING: Final[bytes] = b"PING"
PONG: Final[bytes] = b"PONG"
OK: Final[bytes] = b"+OK"
ERR: Final[bytes] = b"-ERR"

PONG_MESSAGE = PONG + utils.CRLF
OK_MESSAGE = OK + utils.CRLF

MESSAGES: Final[list[bytes]] = [
    INFO,
    MSG,
    HMSG,
    CONNECT,
    PUB,
    HPUB,
    SUB,
    UNSUB,
    PING,
    PONG,
    OK,
    ERR
]



class Message:
    def __init__(self, subject: str, payload: bytes, reply_to: str = None, headers: dict[str, str] = None):
        self.subject = subject
        self.payload = payload
        self.reply_to = reply_to
        self.headers = headers



from abc import ABC, abstractmethod


class BaseConnection(ABC):
    @abstractmethod
    async def connect(self, host: str, port: int, timeout: float):
        raise NotImplementedError()

    @abstractmethod
    async def read(self, size: int) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    async def readline(self) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    async def readexactly(self, size: int) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def write(self, data: bytes):
        raise NotImplementedError()

    @abstractmethod
    async def drain(self):
        raise NotImplementedError()

    @abstractmethod
    async def close(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    async def closed(self) -> bool:
        raise NotImplementedError()



import asyncio
import logging
from asyncio import Task
from typing import Final

import msgspec.json
from typing_extensions import Self

from . import parser
from .subscription import Subscription
from .types import Server
from ..connection.base import BaseConnection
from ..connection.tcp import TcpConnection
from ..protocol import protocol
from ..protocol.messages.msg import MsgProtocolMessage, HMsgProtocolMessage
from ..protocol.messages.pub import PubProtocolMessage, HPubProtocolMessage
from ..protocol.messages.service import InfoProtocolMessage, ConnectProtocolMessage

DEFAULT_CONNECTION_TIMEOUT: Final[float] = 2
DEFAULT_INFO_TIMEOUT: Final[float] = 2
CLIENT_LANGUAGE: Final[str] = "python3"
CLIENT_VERSION: Final[str] = "0.0.1-alpha"


class NATSClient:
    def __init__(self):
        self._logger: logging.Logger = logging.getLogger(__name__)
        self._connection: BaseConnection | None = None
        self._server: Server | None = None
        self._handler_task: Task | None = None
        self._subscriptions: dict[str, Subscription] = {}

    @property
    def connection(self):
        return self._connection

    @connection.setter
    def connection(self, value: BaseConnection):
        if self._connection is not None and not self._connection.closed:
            raise RuntimeError("Replacing non-closed connection is not allowed")
        self._connection = value

    @property
    def available(self):
        return (
                self._connection is not None
                and
                not self._connection.closed
                and
                self._server is not None
        )

    @property
    def server(self):
        return self._server

    @property
    def logger(self):
        return self._logger

    async def send(self, message: bytes):
        self._connection.write(message)
        await self._connection.drain()

    async def _process_connection_init(self):
        info = msgspec.json.decode(
            (await self.connection.readline()).split(b" ", 1)[1].decode(),
            type=InfoProtocolMessage
        )
        self._server = Server(
            id=info.server_id,
            headers_support=info.headers,
            max_payload=info.max_payload,
            name=info.server_name,
            protocol=info.proto,
            version=info.version
        )
        await self.send(
            ConnectProtocolMessage(
                verbose=False,
                pedantic=True,
                tls_required=False,
                lang=CLIENT_LANGUAGE,
                version=CLIENT_VERSION,
                headers=True
            ).dump()
        )

    async def _process_msg(self, msg: MsgProtocolMessage):
        pass

    async def _process_hmsg(self, msg: HMsgProtocolMessage):
        pass

    async def _handler(self):
        while True:
            line = await self.connection.readline()
            head = line.split(b" ", 1)
            match head[0].strip():
                case protocol.PING:
                    await self.send(protocol.PONG_MESSAGE)
                case protocol.ERR:
                    self.logger.error("Received NATS Error: %s", head[1])
                case protocol.MSG:
                    await self._process_msg(await parser.parse_msg(line, self.connection))
                case protocol.HMSG:
                    await self._process_hmsg(await parser.parse_hmsg(line, self.connection))
                case _:
                    self.logger.warning(f"Received unknown NATS protocol message: %s", line)

    async def connect(self, host: str, port: int) -> Self:
        self.connection = TcpConnection()
        await self.connection.connect(host, port, DEFAULT_CONNECTION_TIMEOUT)
        await self._process_connection_init()
        self._handler_task = asyncio.get_running_loop().create_task(self._handler())
        return self

    async def close(self):
        self._handler_task.cancel()
        await self.connection.close()

    async def publish(self, subject: str, data: bytes = b"", reply: str = None, headers: dict[str, str] = None):
        if self.server.headers_support and headers is not None:
            await self.send(
                HPubProtocolMessage(
                    subject=subject,
                    payload=data,
                    reply_to=reply,
                    headers=headers
                ).dump()
            )
            return
        await self.send(
            PubProtocolMessage(
                subject=subject,
                payload=data,
                reply_to=reply
            ).dump()
        )



from msgspec import Struct

from .base import SerializableProtocolMessage
from .. import utils, protocol


class SubProtocolMessage(Struct):
    """
    NATS protocol message model for SUB message
    """
    subject: str
    sid: str
    queue_group: str | None = None

    def dump(self) -> bytes:
        """
        Dump NATS protocol SUB message to bytes
        :return: NATS protocol SUB message as bytes-encoded string
        """
        head = utils.build_head(
            protocol.SUB,
            self.subject.encode(),
            b"" if self.queue_group is None else self.queue_group.encode(),
            self.sid.encode()
        )
        return head + utils.CRLF


class UnsubProtocolMessage(Struct):
    """
    NATS protocol message model for UNSUB message
    """
    sid: str
    max_msgs: int | None = None

    def dump(self) -> bytes:
        """
        Dump NATS protocol UNSUB message to bytes
        :return: NATS protocol UNSUB message as bytes-encoded string
        """
        head = utils.build_head(
            protocol.UNSUB,
            self.sid.encode(),
            b"" if self.max_msgs is None else str(self.max_msgs).encode()
        )
        return head + utils.CRLF
